/*
 * Copyright 2010 IHE International (http://www.ihe.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.ihe.gazelle.proxy.model.message;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.Basic;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EntityManager;
import javax.persistence.FetchType;
import javax.persistence.Transient;

import net.ihe.gazelle.proxy.model.CommandFieldValues;
import net.ihe.gazelle.proxy.model.DicomUIDValues;
import net.ihe.gazelle.proxy.model.LabelKeywordDescriptionClass;
import net.ihe.gazelle.proxy.netty.ChannelType;
import net.ihe.gazelle.proxy.netty.channel.ProxySide;
import net.ihe.gazelle.proxy.util.Preferences;

import org.apache.commons.httpclient.methods.multipart.FilePartSource;
import org.apache.commons.httpclient.methods.multipart.PartSource;
import org.apache.commons.lang.StringEscapeUtils;
import org.hibernate.annotations.CollectionOfElements;
import org.hibernate.annotations.IndexColumn;
import org.jboss.seam.annotations.Name;

import com.pixelmed.dicom.AttributeList;
import com.pixelmed.dicom.AttributeTag;
import com.pixelmed.dicom.DicomException;
import com.pixelmed.dicom.DicomInputStream;

@Entity
@Name("dicomMessage")
public class DicomMessage extends AbstractMessage implements java.io.Serializable {

	/** Serial ID version of this object */
	private static final long serialVersionUID = 54153409123922L;

	// private static org.apache.log4j.Logger log =
	// org.apache.log4j.Logger.getLogger(DicomMessage.class);

	private static Map<Integer, String> _mapOfCommandFields = null;
	private static Map<String, String> _mapOfDicomUID = null;
	private static Object _mapOfDicomUIDSync = new Object();
	private static Object _mapOfCommandFieldsSync = new Object();

	@Column(name = "dicom_filecommandset")
	@Basic(fetch = FetchType.LAZY)
	private byte[] fileCommandSet;

	@Column(name = "dicom_filetransfertsyntax")
	private String fileTransfertSyntax;

	@CollectionOfElements(targetElement = java.lang.String.class)
	@IndexColumn(name = "dicom_message_id")
	private List<String> attributesForCommandSet;

	@Column(name = "dicom_affectedsopclassuid")
	private String affectedSopClassUID;

	@Column(name = "dicom_requestedsopclassuid")
	private String requestedSopClassUID;

	@Column(name = "dicom_commandfield")
	private String commandField;

	@Transient
	private AttributeList attributeList;

	public ChannelType getChannelType() {
		return ChannelType.DICOM;
	}

	// Constructors

	public static Map<String, String> getMapOfDicomUID(EntityManager entityManager) {
		if (_mapOfDicomUID == null) {
			synchronized (_mapOfDicomUIDSync) {
				if (_mapOfDicomUID == null) {
					List<DicomUIDValues> listOfDicomUIDValues = DicomUIDValues.getDicomUIDValuesList(entityManager);
					Map<String, String> buildMapOfDicomUID = new HashMap<String, String>();

					for (LabelKeywordDescriptionClass dicomValue : listOfDicomUIDValues) {
						buildMapOfDicomUID.put(dicomValue.getKeyword(), dicomValue.getValue());
					}
					_mapOfDicomUID = buildMapOfDicomUID;
				}
			}
		}
		return _mapOfDicomUID;
	}

	public static Map<Integer, String> getMapOfCommandFields(EntityManager entityManager) {
		if (_mapOfCommandFields == null) {
			synchronized (_mapOfCommandFieldsSync) {
				if (_mapOfCommandFields == null) {

					List<CommandFieldValues> listOfCommandFields = CommandFieldValues
							.getCommandFieldList(entityManager);
					Map<Integer, String> buildMapOfCommandFields = new HashMap<Integer, String>();

					for (LabelKeywordDescriptionClass commandFieldValue : listOfCommandFields) {
						buildMapOfCommandFields.put(new Integer(commandFieldValue.getKeyword()),
								commandFieldValue.getValue());
					}

					_mapOfCommandFields = buildMapOfCommandFields;
				}
			}
		}
		return _mapOfCommandFields;
	}

	public List<String> getAttributesForCommandSet() {
		return attributesForCommandSet;
	}

	public void setAttributesForCommandSet(List<String> attributesForCommandSet) {
		this.attributesForCommandSet = attributesForCommandSet;
	}

	public byte[] getFileCommandSet() {
		return fileCommandSet;
	}

	public String getAttributesForCommandSetAsString() {
		if (attributesForCommandSet == null) {
			return "";
		} else {
			StringBuilder sb = new StringBuilder();
			for (String attribute : attributesForCommandSet) {
				sb.append(attribute);
				sb.append("\n");
			}
			return sb.toString();
		}
	}

	@SuppressWarnings("unchecked")
	public void setFileCommandSet(EntityManager entityManager, byte[] fileCommandSet) {
		this.fileCommandSet = fileCommandSet;

		Map<Integer, String> mapOfCommandFields = getMapOfCommandFields(entityManager);
		Map<String, String> mapOfDicomUID = getMapOfDicomUID(entityManager);

		if (fileCommandSet != null) {
			attributeList = new AttributeList();

			try {
				attributeList.read(new DicomInputStream(new ByteArrayInputStream(fileCommandSet)));
				Set<AttributeTag> keys = attributeList.keySet();
				Iterator<AttributeTag> iterator = keys.iterator();

				while (iterator.hasNext()) {
					if (attributesForCommandSet == null)
						attributesForCommandSet = new ArrayList<String>();

					AttributeTag tag = iterator.next();

					// Command field
					if (tag.equals(new AttributeTag(0x0, 0x0100))) {
						try {
							Integer fieldValue = attributeList.get(tag).getIntegerValues()[0];
							// log.info("cmd field = " + fieldValue ) ;
							if (mapOfCommandFields.containsKey(fieldValue)) {
								commandField = mapOfCommandFields.get(fieldValue);
							} else
								commandField = fieldValue.toString();
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					// affected sop class
					if (tag.equals(new AttributeTag(0x0, 0x0002))) {
						try {
							String fieldValue = attributeList.get(tag).getDelimitedStringValuesOrEmptyString();
							// log.info("affected sop fieldValue = " +
							// fieldValue ) ;
							if (fieldValue != null && mapOfDicomUID.containsKey(fieldValue.toString())) {
								affectedSopClassUID = mapOfDicomUID.get(fieldValue);
							} else
								affectedSopClassUID = fieldValue;
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
					// requested sop class
					if (tag.equals(new AttributeTag(0x0, 0x0003))) {
						try {
							String fieldValue = attributeList.get(tag).getDelimitedStringValuesOrEmptyString();
							// log.info("requested sop fieldValue = " +
							// fieldValue ) ;
							if (fieldValue != null && mapOfDicomUID.containsKey(fieldValue)) {
								requestedSopClassUID = mapOfDicomUID.get(fieldValue);
							} else
								requestedSopClassUID = fieldValue;
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
					attributesForCommandSet.add(attributeList.get(tag).toString().replace("\0", ""));
				}

			} catch (IOException e) {
				e.printStackTrace();
			} catch (DicomException e) {
				e.printStackTrace();
			}
		}
	}

	public String getFileTransfertSyntax() {
		return fileTransfertSyntax;
	}

	public void setFileTransfertSyntax(String fileTransfertSyntax) {
		this.fileTransfertSyntax = fileTransfertSyntax;
	}

	public AttributeList getAttributeList() {
		return attributeList;
	}

	public void setAttributeList(AttributeList attributeList) {
		this.attributeList = attributeList;
	}

	public String getAffectedSopClassUID() {
		return affectedSopClassUID;
	}

	public void setAffectedSopClassUID(String affectedSopClassUID) {
		this.affectedSopClassUID = affectedSopClassUID;
	}

	public String getRequestedSopClassUID() {
		return requestedSopClassUID;
	}

	public void setRequestedSopClassUID(String requestedSopClassUID) {
		this.requestedSopClassUID = requestedSopClassUID;
	}

	public String getCommandField() {
		return commandField;
	}

	public void setCommandField(String commandField) {
		this.commandField = commandField;
	}

	public DicomMessage() {
		super();
	}

	public DicomMessage(String fromIP, Integer localPort, Integer proxyPort, String toIP, Integer remotePort,
			Integer requestChannelId, Integer responseChannelId, ProxySide proxySide) {
		super(fromIP, localPort, proxyPort, toIP, remotePort, requestChannelId, responseChannelId, proxySide);
	}

	public static String humanReadableByteCount(long bytes, boolean si) {
		int unit = si ? 1000 : 1024;
		if (bytes < unit)
			return bytes + " B";
		int exp = (int) (Math.log(bytes) / Math.log(unit));
		String pre = (si ? "kMGTPE" : "KMGTPE").charAt(exp - 1) + (si ? "" : "i");
		return String.format("%.1f %sB", bytes / Math.pow(unit, exp), pre);
	}

	@Override
	public String getInfoGUI() {
		StringBuilder sb = new StringBuilder("");
		sb.append("Aff. ").append(affectedSopClassUID).append("\n");
		// sb.append("Req. ").append(requestedSopClassUID).append("\n");
		sb.append("Cmd. ").append(commandField).append("\n");
		int messageLength = getMessageLength();
		if (messageLength > 0) {
			sb.append(humanReadableByteCount(messageLength, true));
		}

		String sToUse = sb.toString();
		sToUse = StringEscapeUtils.escapeHtml(sToUse);
		sToUse = sToUse.replaceAll("\n", "<br \\>");

		return sToUse;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((affectedSopClassUID == null) ? 0 : affectedSopClassUID.hashCode());
		result = prime * result + ((attributeList == null) ? 0 : attributeList.hashCode());
		result = prime * result + ((attributesForCommandSet == null) ? 0 : attributesForCommandSet.hashCode());
		result = prime * result + ((commandField == null) ? 0 : commandField.hashCode());
		result = prime * result + Arrays.hashCode(fileCommandSet);
		result = prime * result + ((fileTransfertSyntax == null) ? 0 : fileTransfertSyntax.hashCode());
		result = prime * result + ((requestedSopClassUID == null) ? 0 : requestedSopClassUID.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!super.equals(obj))
			return false;
		if (getClass() != obj.getClass())
			return false;
		DicomMessage other = (DicomMessage) obj;
		if (affectedSopClassUID == null) {
			if (other.affectedSopClassUID != null)
				return false;
		} else if (!affectedSopClassUID.equals(other.affectedSopClassUID))
			return false;
		if (attributeList == null) {
			if (other.attributeList != null)
				return false;
		} else if (!attributeList.equals(other.attributeList))
			return false;
		if (attributesForCommandSet == null) {
			if (other.attributesForCommandSet != null)
				return false;
		} else if (!attributesForCommandSet.equals(other.attributesForCommandSet))
			return false;
		if (commandField == null) {
			if (other.commandField != null)
				return false;
		} else if (!commandField.equals(other.commandField))
			return false;
		if (!Arrays.equals(fileCommandSet, other.fileCommandSet))
			return false;
		if (fileTransfertSyntax == null) {
			if (other.fileTransfertSyntax != null)
				return false;
		} else if (!fileTransfertSyntax.equals(other.fileTransfertSyntax))
			return false;
		if (requestedSopClassUID == null) {
			if (other.requestedSopClassUID != null)
				return false;
		} else if (!requestedSopClassUID.equals(other.requestedSopClassUID))
			return false;
		return true;
	}

	private File getFile() {
		File storageFolder = new File(Preferences.getProperty("storage.dicom"));
		return new File(storageFolder, getMessageReceivedAsString());
	}

	@Override
	public int getMessageLength() {
		try {
			return (int) getFile().length();
		} catch (Throwable e) {
			return 0;
		}
	}

	@Override
	public InputStream getMessageReceivedStream() {
		try {
			return new FileInputStream(getFile());
		} catch (FileNotFoundException e) {
			return null;
		}
	}

	@Override
	public PartSource getPartSource(String messageType) {
		try {
			return new FilePartSource(messageType, getFile());
		} catch (FileNotFoundException e) {
			return null;
		}
	}

}
